import os
import typing

import numpy
import torch
import torchvision
import matplotlib.pyplot as plt

import cvtda.logging
import cvtda.dumping
import cvtda.neural_network

from .BaseLearner import BaseLearner
from .DiagramsLearner import DiagramsLearner
from .NNLearner import NNLearner
from .SimpleTopologicalLearner import SimpleTopologicalLearner


def learn(
    # Train
    train_images: numpy.ndarray,
    train_features: numpy.ndarray,
    train_labels: numpy.ndarray,
    train_diagrams: typing.List[numpy.ndarray],
    # Test
    test_images: numpy.ndarray,
    test_features: numpy.ndarray,
    test_labels: numpy.ndarray,
    test_diagrams: typing.List[numpy.ndarray],
    # General
    n_jobs: int = 1,
    random_state: int = 42,
    dump_name: typing.Optional[str] = None,
    # Neural networks
    nn_device: torch.device = cvtda.neural_network.default_device,
    nn_batch_size: int = 32,
    nn_learning_rate: float = 1e-3,
    nn_epochs: int = 25,
    nn_margin: float = 0.1,
    nn_latent_dim: int = 256,
    nn_length_before_new_iter: typing.Optional[int] = None,
    nn_base=torchvision.models.resnet34,
):
    """
    Tries 6 face recognition models on the given dataset.
    (1) Uses topological feature vectors with Euclidean distance as the latent space.
    (2) Uses persistence diagrams with the bottleneck distance as the latent space.
    (3) Neural network with only traditional CNN on raw images.
    (4) Neural network with only topological features.
    (5) Neural network with both topological features and traditional CNN.
    (6) Neural network with only persistence diagrams.

    Parameters
    ----------
    train_images : ``numpy.ndarray``
        (size `num_items x width x height x num_channels`) Images of the training set.
    train_features : ``numpy.ndarray``
        (size `num_items x num_features`) Topological features of the training set.
    train_labels : ``numpy.ndarray``
        (size `num_items`) Target class labels for each object of the training set.
    train_diagrams : ``numpy.ndarray``
        (size `num_items x num_diagrams x num_points x 3`) Persistence diagrams of the training set.

    test_images : ``numpy.ndarray``
        (size `num_items x width x height x num_channels`) Images of the test set.
    test_features : ``numpy.ndarray``
        (size `num_items x num_features`) Topological features of the test set.
    test_labels : ``numpy.ndarray``
        (size `num_items`) Target class labels for each object of the test set.
    test_diagrams : ``numpy.ndarray``, optional
        (size `num_items x num_diagrams x num_points x 3`) Persistence diagrams of the test set.

    n_jobs : ``int``, default: ``-1``
        The number of jobs to use for the computation. See :mod:`joblib` for details.
    random_state : ``int``, default ``42``
        The seed to initialize the pseudo random generator.
    dump_name : ``str``, optional, default ``None``
        The root to dump the results in, if dumping is enabled.

    nn_device : ``torch.device``, default `cvtda.neural_network.default_device`
        A :mod:`torch` device to perform the computations on.
    nn_batch_size : ``int``, default `32`
        Batch size to train the neural networks with.
    nn_learning_rate : ``float``, default `1e-3`
        Learning rate to train the neural networks with.
    nn_epochs : ``int``, default `25`
        Number of epochs to train the neural networks for.
    nn_margin : ``float``, default `0.1`
        Margin for the triplet loss function used to train neural networks.
    nn_latent_dim : ``int``, default `256`
        Dimensionality of the latent space produces by the autoencoder.
    nn_length_before_new_iter : ``int``, optional, default None
        Number of batches sampled at each epoch. Defaults to 20 * batch_size if not specified.
    nn_base : ``torch.nn.Module``, default `torchvision.models.resnet34`
        Constructor of the CNN model used as a baseline for comparison.

    Returns
    -------
    ``matplotlib.pyplot.Figure``
        A :mod:`matplotlib` figure with distributions of distances between
        photos of same person and different people for analysis.
    """

    # Create datasets
    nn_train = cvtda.neural_network.Dataset(
        train_images, train_diagrams, train_features, train_labels, n_jobs=n_jobs, device=nn_device
    )
    nn_test = cvtda.neural_network.Dataset(
        test_images, test_diagrams, test_features, test_labels, n_jobs=n_jobs, device=nn_device
    )

    def classify_one(learner: BaseLearner, name: str, display_name: str, ax: plt.Axes):
        cvtda.logging.logger().print(f"Trying {name} - {learner}")
        learner.fit(nn_train, nn_test)  # Train
        ax.set_title(display_name)
        learner.estimate_quality(nn_test, ax)  # Test

    # Initialize models.
    nn_kwargs = dict(
        n_jobs=n_jobs,
        random_state=random_state,
        device=nn_device,
        batch_size=nn_batch_size,
        learning_rate=nn_learning_rate,
        margin=nn_margin,
        latent_dim=nn_latent_dim,
        length_before_new_iter=nn_length_before_new_iter,
    )
    classifiers = [
        SimpleTopologicalLearner(n_jobs=n_jobs),
        DiagramsLearner(n_jobs=n_jobs),
        NNLearner(
            **nn_kwargs, n_epochs=nn_epochs, skip_diagrams=True, skip_images=False, skip_features=True, base=nn_base
        ),
        NNLearner(
            **nn_kwargs, n_epochs=nn_epochs * 2, skip_diagrams=True, skip_images=True, skip_features=False, base=nn_base
        ),
        NNLearner(
            **nn_kwargs, n_epochs=nn_epochs, skip_diagrams=True, skip_images=False, skip_features=False, base=nn_base
        ),
        NNLearner(
            **nn_kwargs,
            n_epochs=nn_epochs // 2,
            skip_diagrams=False,
            skip_images=True,
            skip_features=True,
            base=nn_base,
        ),
    ]
    names = [
        "SimpleTopologicalLearner",
        "DiagramsLearner",
        "NNLearner_images",
        "NNLearner_features",
        "NNLearner_features_images",
        "NNLearner_diagrams",
    ]
    display_names = [
        "Topological features",
        "Persistence diagrams",
        "Baseline model",
        "FC over topological features",
        "Combined neural network",
        "Trainable vectorization",
    ]

    # Run models
    figure, axes = plt.subplots(2, 3, figsize=(12, 5))
    for args in zip(classifiers, names, display_names, axes.flat):
        classify_one(*args)

    # Improve presentation
    handles, labels = axes.flat[0].get_legend_handles_labels()
    figure.legend(handles, labels, loc=(0.35, 0.75))
    figure.tight_layout()

    # Dump results, if needed
    dumper = cvtda.dumping.dumper()
    if (dump_name is not None) and isinstance(dumper, cvtda.dumping.NumpyDumper):
        file = dumper.get_file_name_(cvtda.dumping.dump_name_concat(dump_name, "distributions"))
        os.makedirs(os.path.dirname(file), exist_ok=True)
        figure.savefig(file[:-4] + ".svg")
        figure.savefig(file[:-4] + ".png")
    return figure



import joblib
import typing
import inspect

import cvtda.logging
import cvtda.dumping


def parallel(
    function,
    iterable,
    n_jobs: int = -1,
    return_as: str = "list",
    num_parameters: typing.Optional[str] = None,
):
    """
    Utility wrapper around :obj:`joblib.Parallel` that correctly transfers
    :mod:`cvtda.logging` and :mod:`cvtda.dumping` contexts to the created threads.

    Parameters
    ----------
    function : ``Callable``
        The function to execute.
    iterable : ``Iterable``
        The set of elements to execute the function with. Each element will be passed to `function`.
    n_jobs : ``int``, default: ``-1``
        The number of jobs to use for the computation. See :mod:`joblib` for details.
    return_as : ``str``, default: ``list``
        The return value type of :obj:`joblib.Parallel`. See :obj:`joblib.Parallel` for details.

    Returns
    -------
    ``joblib.Parallel``
        The parallel instance performing the computations.
    """

    def init_worker(logger, dumper):
        cvtda.logging.BaseLogger.current_logger = logger
        cvtda.dumping.BaseDumper.current_dumper = dumper

    initargs = (cvtda.logging.logger(), cvtda.dumping.dumper())
    num_parameters = num_parameters or len(inspect.signature(function).parameters)
    with joblib.parallel_backend(backend="loky", initializer=init_worker, initargs=initargs):
        if num_parameters == 1:
            return joblib.Parallel(return_as=return_as, n_jobs=n_jobs)(
                joblib.delayed(function)(item) for item in iterable
            )
        else:
            return joblib.Parallel(return_as=return_as, n_jobs=n_jobs)(
                joblib.delayed(function)(*item) for item in iterable
            )



"""
Defines 'default_device' for operations with :mod:`torch`.
- If cuda is available, uses cuda.
- Else, if mps is available, uses mps.
- Otherwise, falls back to cpu.
"""

import torch


if torch.cuda.is_available():
    default_device = torch.device("cuda")
elif torch.mps.is_available():
    default_device = torch.device("mps")
else:
    default_device = torch.device("cpu")



from .classify import classify
from .estimate_quality import estimate_quality
from .information_value import calculate_binary_information_value
from .information_value import calculate_information_value
from .information_value import InformationValueFeatureSelector
from .NNClassifier import NNClassifier
from .MultiImageClassifier import MultiImageDataset
from .MultiImageClassifier import MultiImageClassifier

__all__ = [
    "classify",
    "estimate_quality",
    "calculate_binary_information_value",
    "calculate_information_value",
    "InformationValueFeatureSelector",
    "NNClassifier",
    "MultiImageDataset",
    "MultiImageClassifier",
]



import os
import random

import numpy
import torch


def set_random_seed(seed: int = 42):
    """
    Seed the random numbers generators in `random`, `numpy`, and `torch`.

    Parameters
    ----------
    seed : ``int``
        The seed to set. Defaults to `42`.
    """
    random.seed(seed)
    numpy.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    torch.backends.cudnn.deterministic = True



import math
import typing
import dataclasses

import numpy
import joblib
import itertools
import gtda.images
import gtda.homology

import cvtda.utils
import cvtda.logging

from . import utils
import cvtda.dumping
from .interface import TopologicalExtractor
from .DiagramVectorizer import DiagramVectorizer


class FiltrationExtractor(TopologicalExtractor):
    """
    Binarizes the images, computes the filtration, and uses cubical persistence to calculate the features.
    """

    def __init__(
        self,
        filtration_class,
        filtation_kwargs: dict,
        binarizer_threshold: float,
        diagram_settings: DiagramVectorizer.Settings,
        n_jobs: int = -1,
        return_diagrams: bool = False,
        only_get_from_dump: bool = False,
        **kwargs,
    ):
        """
        Parameters
        ----------
        filtration_class : ``Any``
            The filtration class from :mod:`gtda.images` used for filtration.
        filtation_kwargs : ``dict``
            Kwargs to pass to the filtration class constructor.
        binarizer_threshold : ``float``
            Threshold for image binarization.
        diagram_settings : ``DiagramVectorizer.Settings``
            Settings for diagram vectorization.
        n_jobs : ``int``, default: ``-1``
            The number of jobs to use for the computation. See :mod:`joblib` for details.
        return_diagrams : ``bool``, default `False`
            If true, :meth:`transform` returns raw persistence diagrams rather then final features.
        only_get_from_dump : ``bool``, default `False`
            If true, all results will be obtained from dump, and no computations will be performed.
        """
        super().__init__(
            enabled=True,
            filtration_class=filtration_class,
            filtation_kwargs=filtation_kwargs,
            binarizer_threshold=binarizer_threshold,
            vectorizer_settings=diagram_settings,
            supports_rgb=False,
            n_jobs=n_jobs,
            return_diagrams=return_diagrams,
            diagram_settings=diagram_settings,
            only_get_from_dump=only_get_from_dump,
            **kwargs,
        )

        self.binarizer_ = gtda.images.Binarizer(threshold=binarizer_threshold, n_jobs=self.n_jobs_)
        self.filtration_ = filtration_class(**filtation_kwargs, n_jobs=self.n_jobs_)
        self.persistence_ = None

    def get_diagrams_(self, images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None):
        cvtda.logging.logger().print(
            f"FiltrationExtractor: processing {dump_name}, do_fit = {do_fit}, filtration = {self.filtration_}"
        )

        if do_fit and (self.persistence_ is None):
            dims = list(range(len(images.shape) - 1))
            self.persistence_ = gtda.homology.CubicalPersistence(homology_dimensions=dims, n_jobs=self.n_jobs_)

        # Binarization
        bin_images = utils.process_iter(self.binarizer_, images, do_fit)
        assert bin_images.shape == images.shape

        # Filtration
        filtrations = utils.process_iter(self.filtration_, bin_images, do_fit)
        assert filtrations.shape == images.shape

        # Persistence
        return utils.process_iter_dump(self.persistence_, filtrations, do_fit, self.diagrams_dump_(dump_name))


class FiltrationsExtractor(cvtda.utils.FeatureExtractorBase):
    """
    Extracts features from grayscale images with a set of
    :class:`FiltrationExtractor` instances with different hyperparameters.
    """

    @dataclasses.dataclass(frozen=True)
    class Settings:
        """
        Attributes
        ----------
        binarizer_thresholds : ``list[float]``
            Threshold for image binarization.
        height_directions : ``list[float]``, optional, default None
            Directions for height filtrations. See :class:`gtda.images.HeightFiltration` for details.
            If not specified, the directions will be initialized dynamically on the training set
            with a 45 degree angle between any two adjacent directions.
        num_radial : ``int``, default 4
            The number of points to sample over each dimension for radial filtration centers.
            The points are samples uniformly using :func:`cvtda.utils.spread_points`.
            See :class:`gtda.images.RadialFiltration` for details on the filtration.
        dilation : ``bool``, default False
            Whether to use a dilation filtration. See :class:`gtda.images.DilationFiltration` for details.
        dilation : ``bool``, default False
            Whether to use a dilation filtration. See :class:`gtda.images.DilationFiltration` for details.
        erosion : ``bool``, default False
            Whether to use an erosion filtration. See :class:`gtda.images.ErosionFiltration` for details.
        signed_distance : ``bool``, default False
            Whether to use a signed distance filtration. See :class:`gtda.images.SignedDistanceFiltration` for details.
        density_radiuses : ``Iterable[int]``, default []
            Radiuses for density filtrations. See :class:`gtda.images.DensityFiltration` for details.
        vectorizer : ``DiagramVectorizer.Settings``
            Settings for diagram vectorization.
        """

        binarizer_thresholds: typing.List[float] = dataclasses.field(default_factory=lambda: [0.25, 0.5, 0.75])

        height_directions: typing.Optional[typing.Iterable[typing.Tuple[float, float]]] = None
        num_radial: int = 4
        dilation: bool = False
        erosion: bool = False
        signed_distance: bool = False
        density_radiuses: typing.Iterable[int] = dataclasses.field(default_factory=lambda: [])

        vectorizer: DiagramVectorizer.Settings = DiagramVectorizer.Settings()

    PRESETS = cvtda.utils.FeatureExtractorBase.Presets(
        full=Settings(
            binarizer_thresholds=[0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9],
            height_directions=None,
            num_radial=4,
            dilation=True,
            erosion=True,
            signed_distance=True,
            density_radiuses=[1, 3],
            vectorizer=DiagramVectorizer.PRESETS.full,
        ),
        reduced=Settings(
            binarizer_thresholds=[0.25, 0.5, 0.75],
            height_directions=None,
            num_radial=4,
            dilation=False,
            erosion=False,
            signed_distance=False,
            density_radiuses=[],
            vectorizer=DiagramVectorizer.PRESETS.reduced,
        ),
        quick=Settings(
            binarizer_thresholds=[0.5],
            height_directions=[(-1, -1), (1, -1), (-1, 1), (1, 1)],
            num_radial=2,
            dilation=False,
            erosion=False,
            signed_distance=False,
            density_radiuses=[],
            vectorizer=DiagramVectorizer.PRESETS.quick,
        ),
    )

    def __init__(
        self,
        settings: Settings,
        n_jobs: int = -1,
        return_diagrams: bool = False,
        only_get_from_dump: bool = False,
    ):
        self.fitted_ = False
        self.n_jobs_ = n_jobs
        self.settings_ = settings
        self.return_diagrams_ = return_diagrams
        self.only_get_from_dump_ = only_get_from_dump
        self.filtration_extractors_: typing.List[typing.Tuple[FiltrationExtractor, str, str]] = []

    def feature_names(self) -> typing.List[str]:
        feature_names = []
        for extractor, _, readable_name in self.filtration_extractors_:
            feature_names.extend(self.nest_feature_names(readable_name, extractor.feature_names()))
        return feature_names

    def fit(self, images: numpy.ndarray, dump_name: typing.Optional[str] = None):
        self.fit_transform(images, dump_name)
        return self

    def transform(self, images: numpy.ndarray, dump_name: typing.Optional[str] = None) -> numpy.ndarray:
        assert self.fitted_ is True, "fit() must be called before transform()"
        cvtda.logging.logger().print("Applying filtrations")
        return self.do_work_(images, do_fit=False, dump_name=dump_name)

    def fit_transform(self, images: numpy.ndarray, dump_name: typing.Optional[str] = None) -> numpy.ndarray:
        assert len(images.shape) >= 3, f"{len(images.shape) - 1}d images are not supported"
        cvtda.logging.logger().print("Fitting filtrations")

        shape = images.shape
        if (len(shape) == 4) and (shape[-1] == 3):
            shape = shape[:-1]
        self.fill_filtrations_(*shape[1:])

        result = self.do_work_(images, do_fit=True, dump_name=dump_name)
        self.fitted_ = True
        return result

    def do_work_(self, images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None) -> numpy.ndarray:
        """
        Performs the actual computations for a batch of images.

        Parameters
        ----------
        images : ``numpy.ndarray``
            (size `num_items x width x height x num_channels`) Input images in grayscale or rgb format.
        do_fit : ``bool``
            If True, will also fit the extractor.
        dump_name : ``str``, optional
            The root to dump the results in, if dumping is enabled.

        Returns
        -------
        ``numpy.ndarray`` or ``list[numpy.ndarray]``
            The result.
        """

        def do_work_one(extractor: FiltrationExtractor, name, readable_name):
            # Do work with one filtration
            extractor_dump_name = cvtda.dumping.dump_name_concat(dump_name, name)
            with cvtda.logging.DevNullLogger():
                result = utils.process_iter(extractor, images, do_fit, dump_name=extractor_dump_name)
                return ((extractor, name, readable_name), result) if do_fit else result

        parallel = cvtda.utils.parallel(
            do_work_one, self.filtration_extractors_, return_as="generator", n_jobs=self.outer_n_jobs_
        )
        outputs = list(cvtda.logging.logger().pbar(parallel, total=len(self.filtration_extractors_)))
        if do_fit:
            self.filtration_extractors_ = [output[0] for output in outputs]
            outputs = [output[1] for output in outputs]
        result = utils.hstack(outputs, not self.return_diagrams_)
        if not self.return_diagrams_ and not self.only_get_from_dump_:
            assert result.shape == (len(images), len(self.feature_names()))
        return result

    def fill_filtrations_(self, *shape: typing.List[int]):
        self.inner_n_jobs_ = 1
        self.do_fill_filtrations_(*shape)

        # We try to utilize the CPU cores as efficiently as possible
        # and dynamically balance the number of filtrations processed in parallel
        # and the parallelization level used within each filtration.
        n_jobs = joblib.effective_n_jobs(self.n_jobs_)
        if len(self.filtration_extractors_) > n_jobs * 3:
            self.outer_n_jobs_ = -1
            self.inner_n_jobs_ = 1
        else:
            self.outer_n_jobs_ = math.gcd(n_jobs, len(self.filtration_extractors_))
            self.inner_n_jobs_ = n_jobs // self.outer_n_jobs_
        self.filtration_extractors_ = []
        self.do_fill_filtrations_(*shape)

    def do_fill_filtrations_(self, *shape: typing.List[int]):
        self.filtration_extractors_ = []
        for binarizer_threshold in self.settings_.binarizer_thresholds:
            self.add_height_filtrations_(binarizer_threshold, *shape)
            self.add_radial_filtrations_(binarizer_threshold, *shape)
            self.add_dilation_filtrations_(binarizer_threshold)
            self.add_erosion_filtrations_(binarizer_threshold)
            self.add_signed_distance_filtrations_(binarizer_threshold)
            self.add_density_filtrations_(binarizer_threshold)

    def make_filtration_(self, filtration_class, filtation_kwargs: dict, binarizer_threshold: float):
        return FiltrationExtractor(
            filtration_class,
            filtation_kwargs,
            binarizer_threshold,
            self.settings_.vectorizer,
            n_jobs=self.inner_n_jobs_,
            return_diagrams=self.return_diagrams_,
            only_get_from_dump=self.only_get_from_dump_,
        )

    def add_height_filtrations_(self, binarizer_threshold: float, *shape: typing.List[int]):
        directions = []
        if self.settings_.height_directions is None:
            # Initialize directions
            directions = list(itertools.product(*([[-1, 0, 1]] * len(shape))))
            directions = filter(lambda item: not all(i == 0 for i in item), directions)
            directions = list(directions)
        else:
            directions = self.settings_.height_directions

        for direction in directions:
            self.filtration_extractors_.append(
                (
                    self.make_filtration_(
                        gtda.images.HeightFiltration, {"direction": numpy.array(direction)}, binarizer_threshold
                    ),
                    f"{int(binarizer_threshold * 10)}/HeightFiltration_{direction[0]}_{direction[1]}",
                    f"HeightFiltration with d = ({direction[0]}, {direction[1]}), bin. thr. = 0.{int(binarizer_threshold * 10)}",
                )
            )

    def add_radial_filtrations_(self, binarizer_threshold: float, *shape: typing.List[int]):
        points = [cvtda.utils.spread_points(coord, self.settings_.num_radial) for coord in shape]
        for center in list(itertools.product(*points)):
            self.filtration_extractors_.append(
                (
                    self.make_filtration_(
                        gtda.images.RadialFiltration, {"center": numpy.array(center)}, binarizer_threshold
                    ),
                    f"{int(binarizer_threshold * 10)}/RadialFiltration_{center[0]}_{center[1]}",
                    f"RadialFiltration with c = ({center[0]}, {center[1]}), bin. thr. = 0.{int(binarizer_threshold * 10)}",
                )
            )

    def add_dilation_filtrations_(self, binarizer_threshold: float):
        if not self.settings_.dilation:
            return
        self.filtration_extractors_.append(
            (
                self.make_filtration_(gtda.images.DilationFiltration, {}, binarizer_threshold),
                f"{int(binarizer_threshold * 10)}/DilationFiltration",
                f"DilationFiltration, bin. thr. = 0.{int(binarizer_threshold * 10)}",
            )
        )

    def add_erosion_filtrations_(self, binarizer_threshold: float):
        if not self.settings_.erosion:
            return
        self.filtration_extractors_.append(
            (
                self.make_filtration_(gtda.images.ErosionFiltration, {}, binarizer_threshold),
                f"{int(binarizer_threshold * 10)}/ErosionFiltration",
                f"ErosionFiltration, bin. thr. = 0.{int(binarizer_threshold * 10)}",
            )
        )

    def add_signed_distance_filtrations_(self, binarizer_threshold: float):
        if not self.settings_.signed_distance:
            return
        self.filtration_extractors_.append(
            (
                self.make_filtration_(gtda.images.SignedDistanceFiltration, {}, binarizer_threshold),
                f"{int(binarizer_threshold * 10)}/SignedDistanceFiltration",
                f"SignedDistanceFiltration, bin. thr. = 0.{int(binarizer_threshold * 10)}",
            )
        )

    def add_density_filtrations_(self, binarizer_threshold: float):
        for radius in self.settings_.density_radiuses:
            self.filtration_extractors_.append(
                (
                    self.make_filtration_(gtda.images.DensityFiltration, {"radius": radius}, binarizer_threshold),
                    f"{int(binarizer_threshold * 10)}/DensityFiltration_{radius}",
                    f"DensityFiltration with r = {radius}, bin. thr. = 0.{int(binarizer_threshold * 10)}",
                )
            )



import typing

import numpy
import torch
import pandas

import cvtda.logging
import cvtda.dumping
import cvtda.neural_network

from .Dataset import Dataset
from .MiniUnet import MiniUnet
from .estimate_quality import estimate_quality


def segment(
    # Train
    train_images: numpy.ndarray,
    train_features: numpy.ndarray,
    train_masks: numpy.ndarray,
    # Test
    test_images: numpy.ndarray,
    test_features: numpy.ndarray,
    test_masks: numpy.ndarray,
    # General
    random_state: int = 42,
    dump_name: typing.Optional[str] = None,
    only_get_from_dump: bool = False,
    # Neural networks
    device: torch.device = cvtda.neural_network.default_device,
    batch_size: int = 64,
    learning_rate: float = 1e-3,
    n_epochs: int = 100,
    remove_cross_maps: bool = False,
):
    """
    Tries 4 UNet-based segmentation models on the given dataset.
    (1) Random predictions for baseline.
    (2) Only topological features.
    (3) Only traditional CNN on raw images.
    (4) Both topological features and traditional CNN.

    Parameters
    ----------
    train_images : ``numpy.ndarray``
        (size `num_items x width x height x num_channels`) Images of the training set.
    train_features : ``numpy.ndarray``
        (size `num_items x num_features`) Topological features of the training set.
    train_masks : ``numpy.ndarray``
        (size `num_items`) Target segmentation masks of the training set.

    test_images : ``numpy.ndarray``
        (size `num_items x width x height x num_channels`) Images of the test set.
    test_features : ``numpy.ndarray``
        (size `num_items x num_features`) Topological features of the test set.
    test_masks : ``numpy.ndarray``
        (size `num_items`) Target segmentation masks of the test set.

    random_state : ``int``, default ``42``
        The seed to initialize the pseudo random generator.
    dump_name : ``str``, optional, default ``None``
        The root to dump the results in, if dumping is enabled.
    only_get_from_dump : ``bool``, default `False`
        If true, all results will be obtained from dump, and no computations will be performed.

    device : ``torch.device``, default `cvtda.neural_network.default_device`
        A :mod:`torch` device to perform the computations on.
    batch_size : ``int``, default `64`
        Batch size to train the neural networks with.
    learning_rate : ``float``, default `1e-3`
        Learning rate to train the neural networks with.
    n_epochs : ``int``, default `100`
        Number of epochs to train the neural networks for.

    Returns
    -------
    ``pandas.DataFrame``
        A data frame with quality metrics of each model. See :func:`estimate_quality` for details.
    """
    nn_train = Dataset(train_images, train_features, train_masks)
    nn_test = Dataset(test_images, test_features, test_masks)

    def try_one(model: MiniUnet, name: str, display_name: str):
        cvtda.logging.logger().print(f"Trying {name} - {model}")

        dumper = cvtda.dumping.dumper()
        model_dump_name = cvtda.dumping.dump_name_concat(dump_name, name)
        if only_get_from_dump or dumper.has_dump(model_dump_name):
            # Get from dump, if available
            y_pred_proba = dumper.get_dump(model_dump_name)
        else:
            # Train
            model.fit(nn_train, nn_test)

            # Test
            y_pred_proba = model.predict_proba(nn_test)

            # Dump if requested
            if model_dump_name is not None:
                dumper.save_dump(y_pred_proba, model_dump_name)

        # Calculate quality and return
        result = {"model": display_name, **estimate_quality(y_pred_proba, test_masks)}
        cvtda.logging.logger().print(result)
        return result

    # Initialize models.
    unet_kwargs = dict(
        random_state=random_state,
        device=device,
        batch_size=batch_size,
        learning_rate=learning_rate,
        n_epochs=n_epochs,
        remove_cross_maps=remove_cross_maps,
    )
    models = [
        MiniUnet(**unet_kwargs, with_images=False, with_features=False),
        MiniUnet(**unet_kwargs, with_images=True, with_features=False),
        MiniUnet(**unet_kwargs, with_images=False, with_features=True),
        MiniUnet(**unet_kwargs, with_images=True, with_features=True),
    ]

    names = ["no", "images", "topological", "combined"]
    display_names = ["No features", "Without topological features", "Only topological features", "Combined features"]

    # Run and return
    return pandas.DataFrame([try_one(*args) for args in zip(models, names, display_names)])



import typing

import numpy
import sklearn.base

import cvtda.utils
import cvtda.logging


class DuplicateFeaturesRemover(sklearn.base.TransformerMixin):
    """
    Efficiently removes duplicate features from an object-feature matrix following a divide-and-conquer approach:
    learns from training set in :meth:`fit`, removes the learned features in :meth:`transform`.
    """

    def __init__(
        self,
        n_jobs: int = -1,
        tolerance: float = 1e-8,
        partition_search_batch_size: int = 2000,
        naive_threshold: int = 300,
    ):
        self.fitted_ = False
        self.n_jobs_ = n_jobs

        self.tolerance_ = tolerance

        self.partition_search_batch_size_ = partition_search_batch_size
        self.naive_threshold_ = naive_threshold

    def fit(self, features: numpy.ndarray):
        """
        Finds the duplicate features and saves their indices.

        Parameters
        ----------
        features : ``numpy.ndarray``
            (size `num_items num_features`) The feature matrix.
        """
        self.non_duplicates_ = self.analyze_columns_(features)[0]
        self.fitted_ = True
        return self

    def transform(self, features: numpy.ndarray) -> numpy.ndarray:
        """
        Removes the duplicate features, as found in :meth:`fit`, from the input.

        Parameters
        ----------
        features : ``numpy.ndarray``
            (size `num_items x num_features`) The feature matrix.

        Returns
        -------
        ``numpy.ndarray``
            The feature matrix with duplicate features removed.
        """
        assert self.fitted_ is True, "fit() must be called before transform()"
        return features[:, self.non_duplicates_]

    def naive_find_duplicate_columns_(self, features: numpy.ndarray) -> set:
        """
        Naive O(N^2) implementation useful for datasets with a small number of features.

        Parameters
        ----------
        features : ``numpy.ndarray``
            (size `num_items x num_features`) The feature matrix.

        Returns
        -------
        ``set[int]``
            The indexes of duplicate features.
        """
        duplicates = set()
        features = features.transpose()
        for i in range(features.shape[0]):
            if i in duplicates:
                continue
            comp = numpy.abs(features - features[i]) < self.tolerance_
            comp = numpy.sum(comp, axis=1)
            comp = numpy.where(comp == features.shape[1])[0]
            for j in comp[1:]:
                duplicates.add(j)
        return duplicates

    def analyze_columns_(self, features: numpy.ndarray, force_naive: bool = False, depth: int = 0) -> typing.List[int]:
        """
        Divide-and-conquer implementation that works by partitioning the features
        into groups with possible duplicates by a value of the feature for one object.

        Parameters
        ----------
        features : ``numpy.ndarray``
            (size `num_items x num_features`) The feature matrix.

        Returns
        -------
        ``list[int]``
            The indexes of non-duplicate features.
        """
        if (features.shape[1] < self.naive_threshold_) or force_naive:
            # If a small number of features is left, fallback to a naive O(N^2).
            if force_naive:
                cvtda.logging.logger().print(f"Naive check for {features.shape[1]} features at depth {depth}")
            duplicates = self.naive_find_duplicate_columns_(features)
            return list(set(range(features.shape[1])) - duplicates), list(duplicates)

        partition_by = self.find_best_partition_(features)
        partition_item = features[partition_by, :]
        duplicates = set()

        prev_value = numpy.min(partition_item) - 5 * self.tolerance_
        pbar = cvtda.logging.logger().pbar(numpy.unique(partition_item))
        for partition_value in pbar:
            if partition_value - prev_value < self.tolerance_:
                continue

            partition_idxs = numpy.where(numpy.abs(partition_item - partition_value) < self.tolerance_)[0]
            partition_idxs = numpy.setdiff1d(partition_idxs, numpy.array(list(duplicates)), assume_unique=True)

            cvtda.logging.logger().set_pbar_postfix(
                pbar,
                {"partition_by": partition_by, "num_features": len(partition_idxs), "duplicates": len(duplicates)},
            )

            is_bad_partition = len(partition_idxs) == features.shape[1]
            with cvtda.logging.DevNullLogger():
                _, partition_duplicates = self.analyze_columns_(
                    features[:, partition_idxs],
                    force_naive=is_bad_partition,
                    depth=depth + 1,
                )

            for item in partition_idxs[list(partition_duplicates)]:
                duplicates.add(item)
            prev_value = partition_value

        cvtda.logging.logger().print(f"Found {len(duplicates)} duplicates")
        return list(set(range(features.shape[1])) - duplicates), list(duplicates)

    def find_best_partition_(self, features: numpy.ndarray) -> int:
        def bulk_biggest_partition_(items: numpy.ndarray) -> list:
            return [numpy.max(numpy.unique(item, return_counts=True)[1]) for item in items]

        partitions = [
            features[batch_start : batch_start + self.partition_search_batch_size_]
            for batch_start in range(0, features.shape[0], self.partition_search_batch_size_)
        ]
        partitions = cvtda.utils.parallel(bulk_biggest_partition_, partitions, return_as="list", n_jobs=self.n_jobs_)
        return numpy.argmin(numpy.hstack(partitions))



import numpy

import skimage.color

import cvtda.utils
import cvtda.logging


def rgb2gray(images: numpy.ndarray, n_jobs: int = -1) -> numpy.ndarray:
    """
    Converts a set of images in RGB color space to grayscale.
    See :func:`skimage.color.rgb2gray` for details.

    Parameters
    ----------
    images : ``numpy.ndarray``
        (size `num_items x width x height x 3`) Input images in RGB format.
    n_jobs : ``int``, default: ``-1``
        The number of jobs to use for the computation. See :mod:`joblib` for details.

    Returns
    -------
    ``numpy.ndarray``
        (size `num_items x width x height`) The images in grayscale.
    """
    return numpy.stack(
        cvtda.utils.parallel(
            skimage.color.rgb2gray,
            cvtda.logging.logger().pbar(images, desc="rgb2gray"),
            n_jobs=n_jobs,
            return_as="list",
            num_parameters=1,
        )
    )



import os
import typing

import numpy
import torch
import pandas
import xgboost
import catboost
import torchvision
import sklearn.base
import sklearn.ensemble
import sklearn.neighbors
import matplotlib.pyplot as plt

import cvtda.logging
import cvtda.dumping
import cvtda.neural_network

from .NNClassifier import NNClassifier
from .estimate_quality import estimate_quality


def classify(
    # Train
    train_images: numpy.ndarray,
    train_features: numpy.ndarray,
    train_labels: numpy.ndarray,
    train_diagrams: typing.Optional[typing.List[numpy.ndarray]],
    # Test
    test_images: numpy.ndarray,
    test_features: numpy.ndarray,
    test_labels: numpy.ndarray,
    test_diagrams: typing.Optional[typing.List[numpy.ndarray]],
    # General
    label_names: typing.Optional[typing.List[str]] = None,
    confusion_matrix_include_values: bool = True,
    n_jobs: int = -1,
    random_state: int = 42,
    dump_name: typing.Optional[str] = None,
    only_get_from_dump: bool = False,
    # KNN
    knn_neighbours: int = 50,
    # Random forest
    random_forest_estimators: int = 100,
    # Neural networks
    nn_device: torch.device = cvtda.neural_network.default_device,
    nn_batch_size: int = 128,
    nn_learning_rate: float = 1e-3,
    nn_epochs: int = 20,
    nn_base=torchvision.models.resnet34,
    # Gradient boosting
    grad_boost_max_iter: int = 20,
    grad_boost_max_depth: int = 4,
    grad_boost_max_features: float = 0.1,
    # XGBoost
    xgboost_n_classifiers: int = 25,
    xgboost_max_depth: int = 4,
    xgboost_device: str = "gpu",
    # CatBoost
    catboost_iterations: int = 600,
    catboost_depth: int = 4,
    catboost_device: str = ("GPU" if torch.cuda.is_available() else "CPU"),
):
    """
    Tries 9 classification models on the given dataset.
    (1) KNN. See :class:`sklearn.neighbors.KNeighborsClassifier`.
    (2) Random forest. See :class:`sklearn.ensemble.RandomForestClassifier`.
    (3) Gradient boosting. See :class:`sklearn.ensemble.HistGradientBoostingClassifier`.
    (4) CatBoost. See :class:`catboost.CatBoostClassifier`.
    (5) XGBoost. See :class:`xgboost.XGBClassifier`.
    (6) Neural network with only topological features.
    (7) Neural network with only persistence diagrams.
    (8) Neural network with only traditional CNN on raw images.
    (9) Neural network with both topological features and traditional CNN.

    Parameters
    ----------
    train_images : ``numpy.ndarray``
        (size `num_items x width x height x num_channels`) Images of the training set.
    train_features : ``numpy.ndarray``
        (size `num_items x num_features`) Topological features of the training set.
    train_labels : ``numpy.ndarray``
        (size `num_items`) Target class labels for each object of the training set.
    train_diagrams : ``numpy.ndarray``, optional
        (size `num_items x num_diagrams x num_points x 3`) Persistence diagrams of the training set.
        If not provided, model (4) is excluded from the analysis.

    test_images : ``numpy.ndarray``
        (size `num_items x width x height x num_channels`) Images of the test set.
    test_features : ``numpy.ndarray``
        (size `num_items x num_features`) Topological features of the test set.
    test_labels : ``numpy.ndarray``
        (size `num_items`) Target class labels for each object of the test set.
    test_diagrams : ``numpy.ndarray``, optional
        (size `num_items x num_diagrams x num_points x 3`) Persistence diagrams of the test set.
        If not provided, model (4) is excluded from the analysis.

    label_names : ``list[str]``, optional, default None
        Class label names to print in the confusion matrices.
        If not specified, the classes will be labeled sequentially with numbers.
    confusion_matrix_include_values : ``bool``, default True
        Includes values in confusion matrices.
    n_jobs : ``int``, default: ``-1``
        The number of jobs to use for the computation. See :mod:`joblib` for details.
    random_state : ``int``, default ``42``
        The seed to initialize the pseudo random generator.
    dump_name : ``str``, optional, default ``None``
        The root to dump the results in, if dumping is enabled.
    only_get_from_dump : ``bool``, default `False`
        If true, all results will be obtained from dump, and no computations will be performed.

    knn_neighbours : ``int``, default `50`
        Number of neighbors to use in KNN.

    random_forest_estimators : ``int``, default `100`
        The number of trees in the random forest.

    nn_device : ``torch.device``, default `cvtda.neural_network.default_device`
        A :mod:`torch` device to perform the computations on.
    nn_batch_size : ``int``, default `128`
        Batch size to train the neural networks with.
    nn_learning_rate : ``float``, default `1e-3`
        Learning rate to train the neural networks with.
    nn_epochs : ``int``, default `20`
        Number of epochs to train the neural networks for.
    nn_base : ``torch.nn.Module``, default `torchvision.models.resnet34`
        Constructor of the CNN model used as a baseline for comparison.

    grad_boost_max_iter : ``int``, default `20`
        The maximum number of iterations of the boosting process.
    grad_boost_max_depth : ``int``, default `4`
        The maximum depth of each tree.
    grad_boost_max_features : ``float``, default `0.1`
        Proportion of randomly chosen features in each node split.

    xgboost_n_classifiers : ``int``, default `25`
        The number of trees in XGBoost.
    xgboost_max_depth : ``int``, default `4`
        The maximum depth of each tree.
    xgboost_device : ``str``, default `gpu`
        The device to train XGBoost on.

    catboost_iterations : ``int``, default `600`
        The number of trees in CatBoost.
    catboost_depth : ``int``, default `4`
        Depth of a tree in CatBoost.
    catboost_device : ``str``, default `GPU` if available, else `CPU`
        The device to train CatBoost on.

    Returns
    -------
    ``tuple(pandas.DataFrame, matplotlib.pyplot.Figure)``
        (1) A data frame with quality metrics of each model.
        (2) A matplotlib figure with confusion matrices for each model.
        See :func:`estimate_quality` for details.
    """

    without_diagrams = (train_diagrams is None) and (test_diagrams is None)

    # Create datasets
    if (train_images is not None) and (not only_get_from_dump):
        nn_train = cvtda.neural_network.Dataset(
            train_images, train_diagrams, train_features, train_labels, n_jobs=n_jobs, device=nn_device
        )
        nn_test = cvtda.neural_network.Dataset(
            test_images, test_diagrams, test_features, test_labels, n_jobs=n_jobs, device=nn_device
        )

    def classify_one(classifier: sklearn.base.ClassifierMixin, name: str, display_name: str, ax: plt.Axes):
        if without_diagrams and name == "NNClassifier_diagrams":
            cvtda.logging.logger().print(f"Skipping {name} - {classifier}")
            return {}

        cvtda.logging.logger().print(f"Trying {name} - {classifier}")

        dumper = cvtda.dumping.dumper()
        model_dump_name = cvtda.dumping.dump_name_concat(dump_name, name)
        if only_get_from_dump or dumper.has_dump(model_dump_name):
            # Get from dump, if available
            y_pred_proba = dumper.get_dump(model_dump_name)
        else:
            # Train and test
            if isinstance(classifier, NNClassifier):
                # Different inputs for neural networks
                classifier.fit(nn_train, nn_test)
                y_pred_proba = classifier.predict_proba(nn_test)
            else:
                classifier.fit(train_features, train_labels)
                y_pred_proba = classifier.predict_proba(test_features)

            # Dump if requested
            if model_dump_name is not None:
                dumper.save_dump(y_pred_proba, model_dump_name)

        # Calculate quality and return
        ax.set_title(display_name)
        result = {
            "classifier": display_name,
            **estimate_quality(
                y_pred_proba,
                test_labels,
                ax,
                label_names=label_names,
                confusion_matrix_include_values=confusion_matrix_include_values,
            ),
        }
        cvtda.logging.logger().print(result)
        return result

    # Initialize models.
    classifiers = [
        sklearn.neighbors.KNeighborsClassifier(n_jobs=n_jobs, n_neighbors=knn_neighbours),
        sklearn.ensemble.RandomForestClassifier(
            n_estimators=random_forest_estimators, random_state=random_state, n_jobs=n_jobs
        ),
        sklearn.ensemble.HistGradientBoostingClassifier(
            random_state=random_state,
            max_iter=grad_boost_max_iter,
            max_depth=grad_boost_max_depth,
            max_features=grad_boost_max_features,
            verbose=cvtda.logging.logger().verbosity(),
        ),
        catboost.CatBoostClassifier(
            iterations=catboost_iterations,
            depth=catboost_depth,
            random_seed=random_state,
            loss_function="MultiClass",
            devices="0-3",
            task_type=catboost_device,
            verbose=(cvtda.logging.logger().verbosity() != 0),
        ),
        xgboost.XGBClassifier(
            n_jobs=n_jobs, n_estimators=xgboost_n_classifiers, max_depth=xgboost_max_depth, device=xgboost_device
        ),
        NNClassifier(
            random_state=random_state,
            device=nn_device,
            batch_size=nn_batch_size,
            learning_rate=nn_learning_rate * 10,
            n_epochs=nn_epochs * 2,
            skip_diagrams=True,
            skip_images=True,
            skip_features=False,
            base=nn_base,
        ),
        NNClassifier(
            random_state=random_state,
            device=nn_device,
            batch_size=nn_batch_size,
            learning_rate=nn_learning_rate,
            n_epochs=nn_epochs // 2,
            skip_diagrams=False,
            skip_images=True,
            skip_features=True,
            base=nn_base,
        ),
        NNClassifier(
            random_state=random_state,
            device=nn_device,
            batch_size=nn_batch_size,
            learning_rate=nn_learning_rate,
            n_epochs=nn_epochs,
            skip_diagrams=True,
            skip_images=False,
            skip_features=True,
            base=nn_base,
        ),
        NNClassifier(
            random_state=random_state,
            device=nn_device,
            batch_size=nn_batch_size,
            learning_rate=nn_learning_rate,
            n_epochs=nn_epochs,
            skip_diagrams=True,
            skip_images=False,
            skip_features=False,
            base=nn_base,
        ),
    ]
    names = [
        "KNeighborsClassifier",
        "RandomForestClassifier",
        "HistGradientBoostingClassifier",
        "CatBoostClassifier",
        "XGBClassifier",
        "NNClassifier_features",
        "NNClassifier_diagrams",
        "NNClassifier_images",
        "NNClassifier_features_images",
    ]
    display_names = [
        "KNN",
        "Random forest",
        "Histogram-based boosting",
        "CatBoost",
        "XGBoost",
        "FC over topological features",
        "Trainable vectorization",
        "Baseline model",
        "Combined neural network",
    ]

    # Run models
    figure, axes = plt.subplots(3, 3, figsize=(15, 15))
    df = pandas.DataFrame([classify_one(*args) for args in zip(classifiers, names, display_names, axes.flat)])
    figure.tight_layout()

    # Dump confusion matrices, if needed
    dumper = cvtda.dumping.dumper()
    if (dump_name is not None) and isinstance(dumper, cvtda.dumping.NumpyDumper):
        file = dumper.get_file_name_(cvtda.dumping.dump_name_concat(dump_name, "confusion_matrixes"))
        os.makedirs(os.path.dirname(file), exist_ok=True)
        figure.savefig(file[:-4] + ".svg")
        figure.savefig(file[:-4] + ".png")
        df.to_csv(dumper.get_file_name_(cvtda.dumping.dump_name_concat(dump_name, "quality_metrics.csv"))[:-4])
    return df, figure



import typing

import numpy
import torch
import pandas
import torchvision

import cvtda.logging
import cvtda.dumping
import cvtda.neural_network

from .Autoencoder import Autoencoder
from .estimate_quality import estimate_quality


def try_autoencoders(
    # Train
    train_images: numpy.ndarray,
    train_features: numpy.ndarray,
    train_diagrams: typing.Optional[typing.List[numpy.ndarray]],
    # Test
    test_images: numpy.ndarray,
    test_features: numpy.ndarray,
    test_diagrams: typing.Optional[typing.List[numpy.ndarray]],
    # General
    n_jobs: int = -1,
    random_state: int = 42,
    dump_name: typing.Optional[str] = None,
    only_get_from_dump: bool = False,
    # Neural networks
    nn_device: torch.device = cvtda.neural_network.default_device,
    nn_batch_size: int = 128,
    nn_learning_rate: float = 1e-3,
    nn_epochs: int = 20,
    nn_latent_dim: int = 256,
    nn_base=torchvision.models.resnet34,
):
    """
    Tries 4 autoencoder models solving compression on the given dataset.
    (1) Using only topological features.
    (2) Using only traditional CNN on raw images.
    (3) Using both topological features and traditional CNN.
    (4) Using only persistence diagrams.

    Parameters
    ----------
    train_images : ``numpy.ndarray``
        (size `num_items x width x height x num_channels`) Images of the training set.
    train_features : ``numpy.ndarray``
        (size `num_items x num_features`) Topological features of the training set.
    train_diagrams : ``numpy.ndarray``, optional
        (size `num_items x num_diagrams x num_points x 3`) Persistence diagrams of the training set.
        If not provided, model (4) is excluded from the analysis.

    test_images : ``numpy.ndarray``
        (size `num_items x width x height x num_channels`) Images of the test set.
    test_features : ``numpy.ndarray``
        (size `num_items x num_features`) Topological features of the test set.
    test_diagrams : ``numpy.ndarray``, optional
        (size `num_items x num_diagrams x num_points x 3`) Persistence diagrams of the test set.
        If not provided, model (4) is excluded from the analysis.

    n_jobs : ``int``, default: ``-1``
        The number of jobs to use for the computation. See :mod:`joblib` for details.
    random_state : ``int``, default ``42``
        The seed to initialize the pseudo random generator.
    dump_name : ``str``, optional, default ``None``
        The root to dump the results in, if dumping is enabled.
    only_get_from_dump : ``bool``, default `False`
        If true, all results will be obtained from dump, and no computations will be performed.

    nn_device : ``torch.device``, default `cvtda.neural_network.default_device`
        A :mod:`torch` device to perform the computations on.
    nn_batch_size : ``int``, default `128`
        Batch size to train the neural networks with.
    nn_learning_rate : ``float``, default `1e-3`
        Learning rate to train the neural networks with.
    nn_epochs : ``int``, default `20`
        Number of epochs to train the neural networks for.
    nn_latent_dim : ``int``, default `256`
        Dimensionality of the latent space produces by the autoencoder.
    nn_base : ``torch.nn.Module``, default `torchvision.models.resnet34`
        Constructor of the CNN model used as a baseline for comparison.

    Returns
    -------
    ``pandas.DataFrame``
        A data frame with quality metrics of each model. See :func:`estimate_quality` for the list of metrics.
    """

    without_diagrams = (train_diagrams is None) and (test_diagrams is None)

    # Create datasets
    if (train_images is not None) and (not only_get_from_dump):
        nn_train = cvtda.neural_network.Dataset(
            train_images, train_diagrams, train_features, None, n_jobs=n_jobs, device=nn_device
        )
        nn_test = cvtda.neural_network.Dataset(
            test_images, test_diagrams, test_features, None, n_jobs=n_jobs, device=nn_device
        )

    def try_one(model: Autoencoder, name: str, display_name: str):
        if without_diagrams and name == "diagrams":
            cvtda.logging.logger().print(f"Skipping {name} - {model}")
            return {}

        cvtda.logging.logger().print(f"Trying {name} - {model}")

        dumper = cvtda.dumping.dumper()
        encoded_dump_name = cvtda.dumping.dump_name_concat(dump_name, f"{name}_encoded")
        decoded_dump_name = cvtda.dumping.dump_name_concat(dump_name, f"{name}_decoded")
        if only_get_from_dump or dumper.has_dump(decoded_dump_name):
            # Get from dump, if available
            decoded = dumper.get_dump(decoded_dump_name)
        else:
            # Train
            model.fit(nn_train, nn_test)

            # Test
            encoded = model.encode(nn_test)
            decoded = model.decode(encoded)

            # Dump if requested
            if encoded_dump_name is not None:
                dumper.save_dump(encoded, encoded_dump_name)
            if decoded_dump_name is not None:
                dumper.save_dump(decoded, decoded_dump_name)

        # Calculate quality and return
        result = {"model": display_name, **estimate_quality(decoded, test_images)}
        cvtda.logging.logger().print(result)
        return result

    # Initialize models.
    models = [
        Autoencoder(
            random_state=random_state,
            device=nn_device,
            batch_size=nn_batch_size,
            learning_rate=nn_learning_rate * 10,
            n_epochs=nn_epochs * 2,
            latent_dim=nn_latent_dim,
            skip_diagrams=True,
            skip_images=True,
            skip_features=False,
            base=nn_base,
        ),
        Autoencoder(
            random_state=random_state,
            device=nn_device,
            batch_size=nn_batch_size,
            learning_rate=nn_learning_rate,
            n_epochs=nn_epochs,
            latent_dim=nn_latent_dim,
            skip_diagrams=True,
            skip_images=False,
            skip_features=True,
            base=nn_base,
        ),
        Autoencoder(
            random_state=random_state,
            device=nn_device,
            batch_size=nn_batch_size,
            learning_rate=nn_learning_rate,
            n_epochs=nn_epochs,
            latent_dim=nn_latent_dim,
            skip_diagrams=True,
            skip_images=False,
            skip_features=False,
            base=nn_base,
        ),
        Autoencoder(
            random_state=random_state,
            device=nn_device,
            batch_size=nn_batch_size,
            learning_rate=nn_learning_rate,
            n_epochs=nn_epochs // 2,
            latent_dim=nn_latent_dim,
            skip_diagrams=False,
            skip_images=True,
            skip_features=True,
            base=nn_base,
        ),
    ]

    names = ["features", "images", "features_images", "diagrams"]
    display_names = [
        "FC over topological features",
        "Baseline model",
        "Combined neural network",
        "Trainable vectorization",
    ]

    # Run and return
    return pandas.DataFrame([try_one(*args) for args in zip(models, names, display_names)])



import abc
import typing

import numpy
import gtda.diagrams

import cvtda.logging
import cvtda.dumping

from .. import utils
from .Extractor import Extractor
from ..DiagramVectorizer import DiagramVectorizer


class TopologicalExtractor(Extractor, abc.ABC):
    """
    Base class for transforming images into features with persistence diagrams calculated in some way.
    Basically abstracts the vectorization of the diagrams.
    """

    def __init__(
        self,
        enabled: bool,
        vectorizer_settings: DiagramVectorizer.Settings,
        supports_rgb: bool,
        n_jobs: int = -1,
        return_diagrams: bool = False,
        only_get_from_dump: bool = False,
        topo_only_get_from_dump: bool = False,
        **kwargs,
    ):
        """
        Parameters
        ----------
        enabled : ``bool``
            Whether the extractor is enabled in the pipeline.
        vectorizer_settings : ``DiagramVectorizer.Settings``
            Settings for diagram vectorization.
        supports_rgb : ``bool``
            Whether the extractor supports processing RGB images directly.
        n_jobs : ``int``, default: ``-1``
            The number of jobs to use for the computation. See :mod:`joblib` for details.
        return_diagrams : ``bool``, default `False`
            If true, :meth:`transform` returns raw persistence diagrams rather then final features.
        only_get_from_dump : ``bool``, default `False`
            If true, all results will be obtained from dump, and no computations will be performed.
        """
        super().__init__(
            n_jobs=n_jobs,
            return_diagrams=return_diagrams,
            only_get_from_dump=False,
            topo_only_get_from_dump=(topo_only_get_from_dump or only_get_from_dump),
            **kwargs,
        )

        self.topo_only_get_from_dump_ = topo_only_get_from_dump or only_get_from_dump
        self.return_diagrams_ = return_diagrams
        self.supports_rgb_ = supports_rgb

        self.enabled_ = enabled
        self.vectorizer_ = DiagramVectorizer(n_jobs=self.n_jobs_, settings=vectorizer_settings)
        self.scaler_ = gtda.diagrams.Scaler(n_jobs=self.n_jobs_, function=lambda x: numpy.max(x))

    def final_dump_name_(self, dump_name: typing.Optional[str] = None):
        # Disable taking the final dump in Extractor. We need to process reading from dump here (to apply Scaler, e.g.)!
        return None

    def diagrams_dump_(self, dump_name: typing.Optional[str]):
        """
        The name of the dump with the computed persistence diagrams.
        """
        return cvtda.dumping.dump_name_concat(dump_name, "diagrams")

    def force_numpy_(self):
        return not self.return_diagrams_

    def nothing_(self, num_objects: int):
        """
        Produces an empty output for a batch of given size.
        """
        return [] if self.return_diagrams_ else numpy.empty((num_objects, 0))

    def process_rgb_(self, rgb_images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None):
        if not self.supports_rgb_:
            return self.nothing_(len(rgb_images))
        return self.do_work_(rgb_images, do_fit, dump_name)

    def feature_names_rgb_(self) -> typing.List[str]:
        if not self.supports_rgb_ or not self.enabled_:
            return []
        return self.vectorizer_.feature_names()

    def process_gray_(self, gray_images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None):
        return self.do_work_(gray_images, do_fit, dump_name)

    def feature_names_gray_(self) -> typing.List[str]:
        if not self.enabled_:
            return []
        return self.vectorizer_.feature_names()

    def do_work_(self, images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None):
        """
        Performs the actual computations for a batch of images.

        Parameters
        ----------
        images : ``numpy.ndarray``
            (size `num_items x width x height x num_channels`) Input images in grayscale or rgb format.
        do_fit : ``bool``
            If True, will also fit the extractor.
        dump_name : ``str``, optional
            The root to dump the results in, if dumping is enabled.

        Returns
        -------
        ``numpy.ndarray`` or ``list[numpy.ndarray]``
            The result.
        """
        if not self.enabled_:
            # If not enabled, just return empty output for compatibility.
            return self.nothing_(len(images))

        # If the user requests to read all results from dumps, do that.
        # Only diagram scaler needs to be applied in this case.
        if self.topo_only_get_from_dump_:
            if self.return_diagrams_:
                diagrams = cvtda.dumping.dumper().get_dump(self.diagrams_dump_(dump_name))
                cvtda.logging.logger().print("Applying Scaler to persistence diagrams.")
                return utils.process_iter(self.scaler_, diagrams, do_fit)
            else:
                return cvtda.dumping.dumper().get_dump(self.features_dump_(dump_name))

        # Get persistence diagrams
        diagrams = self.get_diagrams_(images, do_fit, dump_name)

        # Reduce persistence diagrams to a common scale
        cvtda.logging.logger().print("Applying Scaler to persistence diagrams.")
        diagrams = numpy.nan_to_num(utils.process_iter(self.scaler_, diagrams, do_fit), 0)
        if self.return_diagrams_:
            return diagrams

        # Vectorize the diagrams into features
        features = utils.process_iter_dump(self.vectorizer_, diagrams, do_fit, self.features_dump_(dump_name))
        assert features.shape == (len(images), len(self.vectorizer_.feature_names()))
        return features

    @abc.abstractmethod
    def get_diagrams_(
        self, images: numpy.ndarray, do_fit: bool, dump_name: typing.Optional[str] = None
    ) -> typing.List[numpy.ndarray]:
        """
        Computes persistence diagrams for a set of images.

        Parameters
        ----------
        images : ``numpy.ndarray``
            (size `num_items x width x height x num_channels`) Input images in grayscale or rgb format.
        do_fit : ``bool``
            If True, will also fit the extractor.
        dump_name : ``str``, optional
            The root to dump the results in, if dumping is enabled.

        Returns
        -------
        ``list[numpy.ndarray]``
            Persistence diagrams.
        """
        pass



import typing

from .base import BaseLogger

T = typing.TypeVar("T")


class DevNullLogger(BaseLogger):
    """
    A logger that suppresses all messages.
    """

    def print(self, data: T, *args) -> None:
        pass

    def verbosity(self) -> int:
        return 0

    def pbar(
        self, data: typing.Iterable[T], total: int = None, desc: typing.Optional[str] = None
    ) -> typing.Iterable[T]:
        return data

    def zip(self, *iterables, desc: typing.Optional[str] = None):
        return zip(*iterables)

    def set_pbar_postfix(self, pbar: typing.Any, data: dict):
        pass



import numpy
import sklearn.base
import matplotlib.pyplot as plt

import cvtda.utils
import cvtda.logging


def calculate_binary_information_value(feature: numpy.ndarray, target: numpy.ndarray, bins: int) -> float:
    """
    Calculates information value for one feature in a binary classification problem.

    Parameters
    ----------
    feature : ``numpy.ndarray``
        (size `num_items`) One feature value for each object.
    target : ``numpy.ndarray``
        (size `num_items`) Target class labels for each object. Must be 0 or 1.
    bins : ``int``, default `10`
        The number of bins to categorize the quantitative features.

    Returns
    -------
    ``float``
        Information value score of the feature.
    """
    # First, we need to transform the quantitive feature into categorical.
    quantiles = numpy.linspace(0, 1, bins + 1)
    bins = numpy.unique(numpy.quantile(feature, quantiles))  # Make bins with quantiles
    bins[0] -= 0.0001
    bins[-1] += 0.0001
    feature = numpy.digitize(feature, bins, right=True)  # Digitize the feature

    # Compute events and non-events
    n, events = [], []
    for bin in range(1, len(bins)):
        mask = feature == bin
        n.append(mask.sum())
        events.append((target * mask).sum())
    n = numpy.array(n)
    events = numpy.array(events)

    non_events = n - events
    events_prc = numpy.maximum(events, 0.5) / events.sum()
    non_events_prc = numpy.maximum(non_events, 0.5) / non_events.sum()

    woe = numpy.log(events_prc / non_events_prc)  # weight of evidence
    iv = woe * (events_prc - non_events_prc)  # information value
    return iv.sum()


def calculate_information_value_one_feature(feature: numpy.ndarray, y_true: numpy.ndarray, bins: int) -> dict:
    """
    Calculates information value for one feature in a multi-class classification problem.
    We define information value for a multi-class classification problem as the average
    information value in the corresponding binary classification problems.

    Parameters
    ----------
    feature : ``numpy.ndarray``
        (size `num_items`) One feature value for each object.
    y_true : ``numpy.ndarray``
        (size `num_items`) Target class labels for each object.
    bins : ``int``, default `10`
        The number of bins to categorize the quantitative features.

    Returns
    -------
    ``float``
        Information value score of the feature.
    """
    IVs = []
    for class_idx in range(numpy.max(y_true)):
        target = (y_true == class_idx).astype(int)
        IVs.append(calculate_binary_information_value(feature, target, bins))
    return numpy.mean(IVs)


def calculate_information_value(
    features: numpy.ndarray, y_true: numpy.ndarray, bins: int = 10, n_jobs: int = -1
) -> numpy.ndarray:
    """
    Calculates information value for multiple features in a multi-class classification problem.

    Parameters
    ----------
    features : ``numpy.ndarray``
        (size `num_items x num_features`) Features for each object.
    y_true : ``numpy.ndarray``
        (size `num_items`) Target class labels for each object.
    bins : ``int``, default `10`
        The number of bins to categorize the quantitative features.
    n_jobs : ``int``, default: ``-1``
        The number of jobs to use for the computation. See :mod:`joblib` for details.

    Returns
    -------
    ``numpy.ndarray``
        (size `num_features`) Information value score for each feature.
    """
    params = [(features[:, idx], y_true, bins) for idx in range(features.shape[1])]
    iv_generator = cvtda.utils.parallel(
        calculate_information_value_one_feature, params, return_as="generator", n_jobs=n_jobs
    )
    pbar = cvtda.logging.logger().pbar(iv_generator, total=features.shape[1], desc="information values")
    return numpy.array(list(pbar))


class InformationValueFeatureSelector(sklearn.base.TransformerMixin):
    """
    Selects best features based on their information value on the training set.

    References
    ----------
    .. [1] David Bridston Osteyee , Irving John Good
            "Information, Weight of Evidence. The Singularity Between Probability Measures and Signal Detection"
            https://doi.org/10.1007/BFb0064126
    """

    def __init__(self, n_jobs: int = -1, bins: int = 10, threshold: float = 0.5):
        self.fitted_ = False
        self.n_jobs_ = n_jobs

        self.bins_ = bins
        self.threshold_ = threshold

    def fit(self, features: numpy.ndarray, target: numpy.ndarray):
        """
        Leaves only good features as selected by :meth:`fit` in the feature matrix.

        Parameters
        ----------
        features : ``numpy.ndarray``
            (size `num_items x num_features`) Features for each object.
        target : ``numpy.ndarray``
            (size `num_items`) Target class labels for each object.
        """
        cvtda.logging.logger().print("Fitting the information value feature selector")
        self.IV_ = calculate_information_value(features, target, bins=self.bins_, n_jobs=self.n_jobs_)
        self.good_features_idx_ = numpy.where(self.IV_ > self.threshold_)[0]

        cvtda.logging.logger().print("Fitting complete")
        self.fitted_ = True
        return self

    def transform(self, features: numpy.ndarray) -> numpy.ndarray:
        """
        Leaves only good features as selected by :meth:`fit` in the feature matrix.

        Parameters
        ----------
        features : ``numpy.ndarray``
            (size `num_items x num_features`) Features for each object.

        Returns
        -------
        ``numpy.ndarray``
            (size `num_items x num_good_features`) Selected features for each object.
        """
        assert self.fitted_ is True, "fit() must be called before transform()"
        return features[:, self.good_features_idx_]

    def hist(self, bins: int = 50) -> plt.Figure:
        """
        Histogram of information value scores for all features seen in :meth:`fit`.

        Parameters
        ----------
        bins : ``int``, default `50`
            Number of bins on the histogram.

        Returns
        -------
        ``matplotlib.pyplot.Figure``
            A :mod:`matplotlib` figure with the histogram.
        """
        assert self.fitted_ is True, "fit() must be called before hist()"
        return plt.hist(self.IV_, bins=bins)



import typing

import numpy
import sklearn.metrics
import matplotlib.pyplot as plt


def estimate_quality(
    y_pred_proba: numpy.ndarray,
    y_true: numpy.ndarray,
    ax: typing.Optional[plt.Axes] = None,
    label_names: typing.Optional[typing.List[str]] = None,
    confusion_matrix_include_values: bool = True,
) -> dict:
    """
    Quality metrics for classification task.

    Parameters
    ----------
    y_pred_proba : ``numpy.ndarray``
        (size `num_items x num_classes`) Predicted probability estimates for each class.
    y_true : ``numpy.ndarray``
        (size `num_items`) Target class labels for each object.
    ax : ``matplotlib.pyplot.Axes``, optional, default None
        A matplotlib axis to draw the confusion matrix in.
    label_names : ``list[str]``, optional, default None
        Class label names to print in the confusion matrix.
        If not specified, the classes will be labeled sequentially with numbers.
    confusion_matrix_include_values : ``bool``, default True
        Includes values in confusion matrix.

    Returns
    -------
    ``dict``
        1. Accuracy. See :func:`sklearn.metrics.accuracy_score`.
        2. AUC-ROC. See :func:`sklearn.metrics.roc_auc_score`.
        3. Precision. See :func:`sklearn.metrics.precision_score`.
        4. Recall. See :func:`sklearn.metrics.recall_score`.
        5. F1-score. See :func:`sklearn.metrics.f1_score`.
        6. Top-k accuracy. See :func:`sklearn.metrics.top_k_accuracy_score`.
    """

    if label_names is None:
        label_names = list(range(y_pred_proba.shape[1]))

    y_pred = numpy.argmax(y_pred_proba, axis=1)
    if ax:
        # Draw confusion matrix, if requested
        sklearn.metrics.ConfusionMatrixDisplay.from_predictions(
            y_true,
            y_pred,
            ax=ax,
            colorbar=False,
            display_labels=label_names,
            include_values=confusion_matrix_include_values,
        )
        ax.set_xticks(ax.get_xticks(), labels=ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")
        ax.set_xlabel(None)
        ax.set_ylabel(None)

    numeric_labels = list(range(y_pred_proba.shape[1]))
    auc_roc_input = y_pred_proba
    multi_class = "ovo"
    average = "macro"

    # Handle binary case
    if y_pred_proba.shape[1] == 2:
        auc_roc_input = y_pred_proba[:, 1]
        multi_class = "raise"
        average = "binary"

    metrics = {
        "Accuracy": sklearn.metrics.accuracy_score(y_true, y_pred),
        "AUC-ROC": sklearn.metrics.roc_auc_score(y_true, auc_roc_input, multi_class=multi_class, labels=numeric_labels),
        "Precision": sklearn.metrics.precision_score(y_true, y_pred, average=average, zero_division=0),
        "Recall": sklearn.metrics.recall_score(y_true, y_pred, average=average),
        "F1-score": sklearn.metrics.f1_score(y_true, y_pred, average=average),
    }
    for k in [2, 3, 5, 7, 9, 15, 30]:
        if y_pred_proba.shape[1] > k:
            metrics[f"TOP-{k} Accuracy"] = sklearn.metrics.top_k_accuracy_score(
                y_true, y_pred_proba, k=k, labels=numeric_labels
            )
    return metrics



from .DiagramVectorizer import DiagramVectorizer
from .FeatureExtractor import FeatureExtractor
from .FiltrationsExtractor import FiltrationExtractor
from .FiltrationsExtractor import FiltrationsExtractor
from .GeometryExtractor import GeometryExtractor
from .GreyscaleExtractor import GreyscaleExtractor
from .PointCloudsExtractor import PointCloudsExtractor

__all__ = [
    "DiagramVectorizer",
    "FeatureExtractor",
    "FiltrationExtractor",
    "FiltrationsExtractor",
    "GeometryExtractor",
    "GreyscaleExtractor",
    "PointCloudsExtractor",
]



import typing

import numpy


def deduce_depth(data):
    if (isinstance(data, numpy.ndarray) or isinstance(data, list)) and len(data) > 0:
        return deduce_depth(data[0]) + 1
    return 0


def hstack(data: typing.List[numpy.ndarray], force_numpy: bool):
    """
    Stack a list of numpy arrays horizontally with appropriate handling of dimension mismatch.

    Parameters
    ----------
    data : ``list[numpy.ndarray]``
        The data to process.
    force_numpy : ``bool``
        If True, the result will be forcefully converted to numpy.

    Returns
    -------
    ``numpy.ndarray`` or ``list[numpy.ndarray]``
        The stacked data.
    """
    if force_numpy:
        return numpy.hstack(data)
    else:
        data = [item for item in data if len(item) > 0]
        if len(data) == 0:
            return []
        assert len(set([len(item) for item in data])) == 1

        # Unsqueeze the data, if it is squeezed for some reason.
        for i in range(len(data)):
            if deduce_depth(data[i]) == 3:
                data[i] = numpy.expand_dims(data[i], 1)

            assert deduce_depth(data[i]) == 4, f"Bad depth: {deduce_depth(data[i])}"

        return [sum([list(item[i]) for item in data], []) for i in range(len(data[0]))]



import numpy
import sklearn.base


def process_iter(transformer: sklearn.base.TransformerMixin, data: numpy.ndarray, do_fit: bool, *args, **kwargs):
    """
    Processes the `data` with the `transformer`, optionally fitting the transformer.

    Parameters
    ----------
    transformer : ``sklearn.base.TransformerMixin``
        The transformer.
    data : ``numpy.ndarray``
        The data to process.
    do_fit : ``bool``
        If True, will also fit the transformer.

    Returns
    -------
    ``Any``
        The output of `transformer.transform`
    """
    if do_fit:
        return transformer.fit_transform(data, *args, **kwargs)
    return transformer.transform(data, *args, **kwargs)



import numpy
import typing

import cvtda.utils
import cvtda.logging


def image2pointcloud(images: numpy.ndarray, n_jobs: int = -1) -> typing.List[numpy.ndarray]:
    """
    Converts a set of images to point clouds.
    Each pixel of the image is converted into a point,
    with the first two dimensions being the coordinates of the pixel,
    and the color values of the pixel for each channel placed in the end.

    Parameters
    ----------
    images : ``numpy.ndarray``
        (size `num_items x width x height x num_channels`) Input images in any format.
    n_jobs : ``int``, default: ``-1``
        The number of jobs to use for the computation. See :mod:`joblib` for details.

    Returns
    -------
    ``numpy.ndarray``
        (size `num_items x (2 + num_channels)`) The resulting point clouds.
    """

    def _impl(image: numpy.ndarray) -> numpy.ndarray:
        width, height = image.shape[0:2]
        x = numpy.indices((width, height))[0]
        y = numpy.indices((width, height))[1]
        return numpy.dstack([x, y, image]).reshape((width * height, -1))

    return cvtda.utils.parallel(
        _impl,
        cvtda.logging.logger().pbar(images, desc="image2pointcloud"),
        n_jobs=n_jobs,
        return_as="list",
    )
